/*****************************************************************
**
**	@(#) ddns.c	(c) Apr 2016 by hznet H.Zuleger
**
**	For zone file parsing it uses the ldns library
**	ldns is Copyright (c) 2005-2008, NLnet Labs
**
**	Dynamic updates and zone transfer requires actually the two
**	BIND tools 'dig' and 'nsupdate' (c) ISC
**
*****************************************************************/

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <ldns/ldns.h>

#include "config.h"

#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
# include <getopt.h>
#endif

#define extern
# include "ddns.h"
#undef extern
#include "misc.h"
#include "readkeyfile.h"
#include "updater.h"

#define LDNS 1

/*****************************************************************
**	local function declaration
*****************************************************************/
static	void	usage (const char *mesg);
static	int	readzone (FILE *fp, FILE *out, const ldns_rdf *origin, ldns_resolver *res, ldns_tsig_credentials *tsig);
static	int	remove_cmd (FILE *fp, long pos, int cmd);

# define	isdelcmd(c)	((c) == '-' || (c) == '<')
# define	isaddcmd(c)	((c) == '+' || (c) == '>')
# define	iscomment(c)	((c) == ';')
# define	iscommand(c)	(isdelcmd(c) || isaddcmd(c))

/*****************************************************************
**	global vars 
*****************************************************************/

extern	int	optopt;
extern	int	opterr;
extern	int	optind;
extern	char	*optarg;


static	const	char	*keyname;
static	const	char	*conffile;
static	int	writeflag = 0;
static	ulong	defttl = 0;
static	int	noexec = 0;
static	ldns_rdf	*origin;
static	tsigkey_t	key;

# define	short_options	":ht:c:k:no:wvV"

#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
static	struct	option	long_options[] = {
	{ "help",		no_argument,		NULL,	'h' },
	{ "version",		no_argument,		NULL,	'V' },
	{ "verbose",		no_argument,		NULL,	'v' },
	{ "ttl",		required_argument,	NULL,	't' },
	{ "keyconf",		required_argument,	NULL,	'c' },
	{ "keyname",		required_argument,	NULL,	'k' },
	{ "noexec",		no_argument,		NULL,	'n' },
	{ "origin",		required_argument,	NULL,	'o' },
	{ "write",		no_argument,		NULL,	'w' },
};
#endif


/*****************************************************************
**	main ()
*****************************************************************/
int	main(int argc, char **argv)
{
	FILE	*fp;
	FILE	*out;
	int	opt_index;
	int	c;
	char	errstr[127+1];
	ldns_tsig_credentials	tsig_cred;
	ldns_resolver	*res;

	version = ZKT_VERSION;
	if ( (progname = strrchr (*argv, '/')) == NULL )
		progname = *argv;
	else
		progname++;

        opterr = 0;     /* prevent the printing of the getopt() build-in error messages */

	keyname = "local-ddns";
	conffile = SESSION_KEY_PATH;

	fp = NULL;

#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
	while ( (c = getopt_long (argc, argv, short_options, long_options, &opt_index)) != -1 )
#else
	while ( (c = getopt (argc, argv, short_options)) != -1 )
#endif
	{
		switch ( c )
		{
		case 'h':
			usage (NULL);
			break;
		case 'V':
			fatal ("%s version %s (compiled with ldns %s)\n", progname, version, ldns_version());
			break;
		case 'v':
			verbose++;
			break;
		case 'n':
			noexec = 1;
			break;
		case 'o':
			origin = ldns_rdf_new_frm_str (LDNS_RDF_TYPE_DNAME, optarg);
			if ( !origin )
			{
				fatal ("error: creating origin from -o %s failed.\n", optarg);
			}
			break;
		case 'k':
			keyname = optarg;
			break;
		case 'c':
			conffile = optarg;
			break;
		case 't':
			ttlfromstr (optarg, &defttl);
			break;
		case 'w':
			writeflag = 1;
			break;
               case ':':
                        snprintf (errstr, sizeof(errstr), "option \"-%c\" requires an argument.\n",
                                                                                optopt);
                        usage (errstr);
                        break;
                case '?':
                        if ( isprint (optopt) )
                                snprintf (errstr, sizeof(errstr), "unknown option \"-%c\".\n", optopt);
                        else
                                snprintf (errstr, sizeof (errstr), "unknown option char \\x%x.\n", optopt);
                        usage (errstr);
                        break;
                default:
                        abort();
		}
	}

	vmesg (2, "Reading key \"%s\" from file \"%s\"\n", keyname, conffile);
	if ( readkeyfile (conffile, &key, keyname) <= 0 )
		fatal ("couldn\'t find key \"%s\" in file \"%s\"\n", keyname, conffile);

	out = stdout;
	if ( noexec == 0 )
	{
#if ! LDNS
		char	updcmd[63+1];

		if ( conffile )
			snprintf (updcmd, sizeof (updcmd), "%s -v -l -k %s", NSUPDATE_PATH, conffile);
		else
			snprintf (updcmd, sizeof (updcmd), "%s -v -l", NSUPDATE_PATH);
		vmesg (2, "Executing \"%s\"\n", updcmd);
		if ( (out = popen (updcmd, "w")) == NULL )
			fatal ("Can't execute %s\n", updcmd);
		if ( verbose > 0 )
			fprintf (out, "debug\n");
		// fprintf (out, "server localhost\n");
		// fprintf (out, "key %s %s %s\n", key.algo, key.name, key.secret);
#else
	tsig_cred.keyname = key.name;
	tsig_cred.algorithm = key.algo;
	tsig_cred.keydata = key.secret;

	res = getlocalresolver (53, &tsig_cred);
#endif
	}

	/* parse remaining args, which are zone file names */
	argc -= optind;
	argv += optind;
	if ( argc == 0 )
	{
		writeflag = 0;
		noexec = 0;
		readzone (stdin, out, origin, res, &tsig_cred);
		exit (0);
	}

	while ( argc-- > 0 )
	{
		if ( (fp = fopen (*argv, writeflag ? "r+": "r")) == NULL )
		{
			snprintf(errstr, sizeof(errstr), "can't open zonefile %s (%s)\n",
								*argv, strerror(errno));
			usage (errstr);
		}

		readzone (fp, out, origin, res, &tsig_cred);

		fclose (fp);

		argv++;
	}

#if ! LDNS
	if ( !noexec )
		pclose (out);
#endif

        exit (0);
}

# define	sopt_usage(mesg, value)	fprintf (stderr, mesg, value)
#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
# define	lopt_usage(mesg, value)	fprintf (stderr, mesg, value)
# define	loptstr(lstr, sstr)	lstr
#else
# define	lopt_usage(mesg, value)
# define	loptstr(lstr, sstr)	sstr
#endif
/*****************************************************************
**	usage ()
*****************************************************************/
static	void	usage (const char *mesg)
{
	if ( mesg && *mesg )
		fprintf (stderr, "%s: %s\n", progname, mesg);
	fprintf (stderr, "usage: %s -h|V\n", progname);
	fprintf (stderr, "usage: %s [-vn] [-t <ttl>] [-o origin] [-c keyfile] [-k keyname] [-w] [zonefile ...]\n", progname);
	fprintf (stderr, "\n");
	fprintf (stderr, "\t-h%s\t print this text\n", loptstr (", --help", "\t"));
	fprintf (stderr, "\t-V%s\t print version information and exit\n", loptstr (", --version", "\t"));

	fprintf (stderr, "\n");
	fprintf (stderr, "\t-v%s\t turn verbose mode on\n", loptstr (", --verbose", "\t"));
	fprintf (stderr, "\t-n%s\t do not send update; just print nsupdate compatible messages to stdout\n", loptstr (", --noexec", "\t"));
        fprintf (stderr, "\t-c file%s", loptstr (", --keyfile=file\n\t", ""));
	fprintf (stderr, "\t\t config file with key definition (default \"%s\")\n", SESSION_KEY_PATH );
        fprintf (stderr, "\t-k keyname%s", loptstr (", --keyname=name\n\t", ""));
	fprintf (stderr, "\t\t name of key used to authenticate dynamic update messages (def: \"%s\")\n", keyname);
        fprintf (stderr, "\t-o zone%s", loptstr (", --origin=zone\n\t", ""));
	fprintf (stderr, "\t\t name of the zone (normally set by $ORIGIN directive in the zonefile)\n");
        fprintf (stderr, "\t-t ttl%s", loptstr (", --ttl=ttlspec\n\t", ""));
	fprintf (stderr, "\t\t overwrite default ttl (default is taken from $TTL directive in zonefile)\n");
	fprintf (stderr, "\t\t\t known ttl units are s(ecs), m(ins), h(ours), d(ays) or w(eeks) \n");

	fprintf (stderr, "\t-w%s\t (re)write zone file and remove update command char\n", loptstr (", --write", "\t"));

	exit (1);
}

/*****************************************************************
**	readzone ()
*****************************************************************/
static	int	readzone (FILE *fp, FILE *out, const ldns_rdf *origin, ldns_resolver *res, ldns_tsig_credentials *tsig)
{
	int	lnr;
	int	c;
	int	nextc;
	int	removelabel = 0;
	long	pos;
	int	soa_seen = 0;
	ldns_rr		*rr;
	ldns_rdf	*my_origin;
	ldns_rdf	*my_prev;
	ldns_rdf	*owner;
	uint32_t	my_ttl;
#if LDNS
	ldns_rr_list	*up_rrlist;

	up_rrlist = ldns_rr_list_new ();
	if ( !up_rrlist )
		fatal ("coudn't create update resource record list\n");
#endif

	assert ( fp != NULL );

	my_origin = NULL;
	my_prev = NULL;
	my_ttl = 7200;
	rr = NULL;

	if ( origin )
	{
		my_origin = ldns_rdf_clone (origin);
		my_prev   = ldns_rdf_clone (origin);
	}

	lnr = 0;
	while ( !feof (fp) )
	{
		pos = ftell (fp);	/* remember beginning of line position */

		/* get the first char of the line */
		c = fgetc (fp);
		if ( iscommand (c) )
		{
			/* command char found.. */
			nextc = fgetc (fp);	/* ... so what follows next ?*/
			ungetc (nextc, fp);
		}
		else
			ungetc (c, fp);	/* push back if it was not a command char */


		vmesg (3, "%4d Start of Line char is '%c'\n", lnr, c);
			/* parse the next rr (or zone directive) */
		switch ( ldns_rr_new_frm_fp_l (&rr, fp, &my_ttl, &my_origin, &my_prev, &lnr) )
		{
		case LDNS_STATUS_OK:
			if ( ldns_rr_get_type (rr) == LDNS_RR_TYPE_SOA )
			{
				if ( soa_seen )
				{
					/* second SOA 
					 * just skip, maybe we want to say
					 * something??? */
					ldns_rr_free (rr);
					error ("warning: second SOA record found in line %4d \n", lnr);
					return 1;
				}
				soa_seen = 1;
				if ( !my_origin )	/* set origin to soa if not specified */
					my_origin = ldns_rdf_clone (ldns_rr_owner (rr));
				vmesg (2, "%4d  %s\n", lnr, "SOA found");
				break;
			}
			/* a normal RR - as sofar the DNS is normal */
			break;
		case LDNS_STATUS_SYNTAX_TTL: /* $TTL directive found */
				/* cmdline ttl given and this is $TTL directive before the first SOA record ? */
			vmesg (2, "%4d  %s %lu\n", lnr, "ttl line", (ulong) my_ttl);
			if ( defttl && !soa_seen )
			{
				my_ttl = defttl;	/* overwrite $TTL with option -t */
				vmesg (2, "      => resetting to command line ttl %lu\n", (ulong) my_ttl);
			}
			continue;
		case LDNS_STATUS_SYNTAX_ORIGIN:	/* the function set always the origin */
			/* this can not be overwritten by option -o */
			vmesg (2, "%4d  %s\n", lnr, "origin line");
			continue;
		case LDNS_STATUS_SYNTAX_INCLUDE:
			vmesg (2, "%4d  %s\n", lnr, "INCLUDE line");
			continue;
		default:
			//fprintf (stderr, "%4d  %s %s \n", lnr, "error", ldns_get_errorstr_by_id (s));
			continue;
		}

		owner = ldns_rr_owner (rr);
		if ( isaddcmd (c) || isdelcmd (c) )
		{
			ldns_rr_list_push_rr (up_rrlist, rr);

			if ( !isspace (nextc) )		/* label after commandchar ? */
				removelabel = 0;	/* reset the marker */
			else if ( removelabel )		/* this line has no label and last rr removed it */
			{
				error ("line %d: RR delete request followed by line w/o label\n", lnr-1);
				error ("line %d: add owner name %s\n", lnr, ldns_rdf2str (owner));
				removelabel = 0;
			}

			/* delete command on a line with owner label? */
			if ( isdelcmd (c) && !isspace (nextc) )
				removelabel = 1;	/* keep in mind that we remove a RR with label */

			if ( noexec || verbose >= 1 )
				fprintf (out, "update %s %s",
						isaddcmd(c) ? "add": "del",
						ldns_rr2str (rr));

			if ( writeflag )
				remove_cmd (fp, pos, c);
		}
		else 
		{
			//fprintf (stderr, "%d '%c' %d\n", lnr, c, removelabel);
			if ( isspace (c) && removelabel )
			{
				error ("line %d: RR delete request followed by line w/o label\n", lnr-1);
				error ("line %d: add owner name %s\n", lnr, ldns_rdf2str (owner));
				removelabel = 0;
			}
			vmesg (1, ";%s", ldns_rr2str (rr));
		}
	}

	if ( noexec )
		fprintf (out, "send\n");
	else
		if ( !sendupdate (res, my_origin, NULL, up_rrlist, tsig) )
			error ("%s: update failed\n", progname);

	if ( up_rrlist )
		ldns_rr_list_deep_free (up_rrlist);

	if (my_origin)
		ldns_rdf_deep_free (my_origin);
	if ( my_prev )
		ldns_rdf_deep_free (my_prev);

	return 1;
}

static	int	remove_cmd (FILE *fp, long pos, int cmd)
{
	int	ret;
	long	currpos;
	char	label[254+1];
	const	char	*p;

	ret = 1;
	currpos = ftell (fp);

	fseek (fp, pos, SEEK_SET);		/* go back to begining of last RR */
	cmd = fgetc (fp);			/* skip command char */
	fscanf (fp, "%254[^ \t]", label);	/* read in owner name */
	fseek (fp, pos, SEEK_SET);		/* go back again */

	vmesg (2, "remove cmd \'%c\' at filepos %lu with label \"%s\" while beeing at %lu\n", cmd, pos, label, currpos);

	if ( cmd == '-' )
	{
		fputc (';', fp);	/* replace '-' with comment char ';' */
		if ( *label )	/* if we remove a line with a label, the next entry is propably dangling */
			;
	}
	else if ( cmd == '+' )
	{
		p = label;
		while ( *p )		/* copy label to start of line */
			fputc (*p++, fp);
		fputc (' ', fp);		/* add a blank */
	}
	else
		ret = 0;

	fseek (fp, currpos, SEEK_SET);	/* go further as though nothing had happend */

	return ret;
}
